@@ -116,6 +116,9 @@ QVariant ZhttpRequestPacket::toVariant() const
 	if(ignorePolicies)
 		obj["ignore-policies"] = true;
 
+	if(trustConnectHost)
+		obj["trust-connect-host"] = true;
+
 	if(ignoreTlsErrors)
 		obj["ignore-tls-errors"] = true;
 
@@ -356,6 +359,15 @@ bool ZhttpRequestPacket::fromVariant(const QVariant &in)
 		ignorePolicies = obj["ignore-policies"].toBool();
 	}
 
+	trustConnectHost = false;
+	if(obj.contains("trust-connect-host"))
+	{
+		if(obj["trust-connect-host"].type() != QVariant::Bool)
+			return false;
+
+		trustConnectHost = obj["trust-connect-host"].toBool();
+	}
+
 	ignoreTlsErrors = false;
 	if(obj.contains("ignore-tls-errors"))
 	{
