@@ -18,14 +18,18 @@
 #include "websocket.h"
 
 #include <assert.h>
+#ifdef HAVE_OPENSSL
+#include <openssl/x509.h>
+#endif
 #include <QUrl>
 #include <QPointer>
 #include <QSslSocket>
 #include "log.h"
 #include "bufferlist.h"
 #include "addressresolver.h"
+#include "verifyhost.h"
 
-#define REJECT_BODY_MAX 100000
+#define RESPONSE_BODY_MAX 100000
 #define MAGIC_STRING "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
 
 static quint16 read16(const quint8 *in)
@@ -319,7 +323,9 @@ public:
 	AddressResolver *resolver;
 	State state;
 	QString connectHost;
+	bool trustConnectHost;
 	bool ignoreTlsErrors;
+	int maxRedirects;
 	int maxFrameSize;
 	QSslSocket *sock;
 	QUrl requestUri;
@@ -330,7 +336,7 @@ public:
 	HttpHeaders responseHeaders;
 	BufferList responseBody;
 	int responseContentLength;
-	bool readingRejectBody;
+	bool readingResponseBody;
 	bool chunked;
 	bool peerClosing;
 	int peerCloseCode;
@@ -344,18 +350,21 @@ public:
 	int inBytes;
 	bool pendingRead;
 	QList<WriteItem> pendingWrites;
+	int followedRedirects;
 
 	Private(WebSocket *_q, QJDnsShared *_dns) :
 		QObject(_q),
 		q(_q),
 		dns(_dns),
 		state(Idle),
+		trustConnectHost(false),
 		ignoreTlsErrors(false),
+		maxRedirects(-1),
 		maxFrameSize(-1),
 		sock(0),
 		responseCode(-1),
 		responseContentLength(-1),
-		readingRejectBody(false),
+		readingResponseBody(false),
 		chunked(false),
 		peerClosing(false),
 		peerCloseCode(-1),
@@ -363,11 +372,12 @@ public:
 		mostSignificantError(ErrorGeneric),
 		inStatusLine(true),
 		inBytes(0),
-		pendingRead(false)
+		pendingRead(false),
+		followedRedirects(0)
 	{
 		resolver = new AddressResolver(dns, this);
-		connect(resolver, SIGNAL(resultsReady(const QList<QHostAddress> &)), SLOT(resolver_resultsReady(const QList<QHostAddress> &)));
-		connect(resolver, SIGNAL(error()), SLOT(resolver_error()));
+		connect(resolver, &AddressResolver::resultsReady, this, &Private::resolver_resultsReady);
+		connect(resolver, &AddressResolver::error, this, &Private::resolver_error);
 	}
 
 	~Private()
@@ -391,10 +401,27 @@ public:
 		requestUri = uri;
 		requestHeaders = headers;
 
+		tryConnect();
+	}
+
+	void tryConnect()
+	{
+		responseCode = -1;
+		responseContentLength = -1;
+		readingResponseBody = false;
+		chunked = false;
+		peerClosing = false;
+		peerCloseCode = -1;
+		errorCondition = ErrorNone;
+		mostSignificantError = ErrorGeneric;
+		inbuf.clear();
+		inStatusLine = true;
+		pendingRead = false;
+
 		if(!connectHost.isEmpty())
 			host = connectHost;
 		else
-			host = uri.host();
+			host = requestUri.host();
 
 		state = Connecting;
 		resolver->start(host);
@@ -553,7 +580,7 @@ public:
 				}
 				else
 				{
-					// we'll read the response body before emitting error
+					// we'll read the response body before acting
 					if(responseHeaders.contains("Content-Length"))
 					{
 						bool ok;
@@ -576,7 +603,7 @@ public:
 					responseHeaders.removeAll("Content-Length");
 					responseHeaders.removeAll("Transfer-Encoding");
 
-					readingRejectBody = true;
+					readingResponseBody = true;
 				}
 			}
 			else
@@ -598,6 +625,46 @@ public:
 		return true;
 	}
 
+	void handleResponse()
+	{
+		if(maxRedirects > 0
+			&& (responseCode == 301 || responseCode == 302 || responseCode == 303 || responseCode == 307 || responseCode == 308)
+			&& responseHeaders.contains("Location"))
+		{
+			QByteArray location = responseHeaders.get("Location");
+
+			log_debug("ws: received redirect response, code=%d location=[%s]", responseCode, location.data());
+
+			if(followedRedirects >= maxRedirects)
+			{
+				log_debug("ws: too many redirects");
+
+				cleanup();
+				state = Idle;
+				errorCondition = ErrorGeneric;
+				emit q->error();
+				return;
+			}
+
+			++followedRedirects;
+
+			requestUri = QUrl::fromEncoded(location);
+
+			cleanup();
+			tryConnect();
+		}
+		else
+		{
+			// force content-length on rejections
+			responseHeaders += HttpHeader("Content-Length", QByteArray::number(responseBody.size()));
+
+			cleanup();
+			state = Idle;
+			errorCondition = ErrorRejected;
+			emit q->error();
+		}
+	}
+
 	// return true if new frame to read
 	bool handleIncomingFrame(bool fin, int opcode, const QByteArray &data)
 	{
@@ -723,7 +790,7 @@ public:
 					break;
 				}
 
-				if(responseBody.size() + size > REJECT_BODY_MAX)
+				if(responseBody.size() + size > RESPONSE_BODY_MAX)
 				{
 					// can't fit the next chunk. we'll stop now
 					eof = true;
@@ -745,7 +812,7 @@ public:
 		{
 			if(!inbuf.isEmpty())
 			{
-				int avail = REJECT_BODY_MAX - responseBody.size();
+				int avail = RESPONSE_BODY_MAX - responseBody.size();
 
 				// don't read more than Content-Length
 				if(responseContentLength != -1)
@@ -755,19 +822,19 @@ public:
 				responseBody += inbuf.mid(0, size);
 				inbuf = inbuf.mid(size);
 
-				assert(responseBody.size() <= REJECT_BODY_MAX);
+				assert(responseBody.size() <= RESPONSE_BODY_MAX);
 			}
 
 			if(responseContentLength != -1)
 			{
 				assert(responseBody.size() <= responseContentLength);
 
-				if(responseBody.size() == responseContentLength || responseBody.size() == REJECT_BODY_MAX)
+				if(responseBody.size() == responseContentLength || responseBody.size() == RESPONSE_BODY_MAX)
 					eof = true;
 			}
 			else
 			{
-				if(responseBody.size() == REJECT_BODY_MAX)
+				if(responseBody.size() == RESPONSE_BODY_MAX)
 					eof = true;
 			}
 
@@ -776,18 +843,7 @@ public:
 		}
 
 		if(eof)
-			respondRejected();
-	}
-
-	void respondRejected()
-	{
-		// force content-length on rejections
-		responseHeaders += HttpHeader("Content-Length", QByteArray::number(responseBody.size()));
-
-		cleanup();
-		state = Idle;
-		errorCondition = ErrorRejected;
-		emit q->error();
+			handleResponse();
 	}
 
 private slots:
@@ -812,12 +868,12 @@ private slots:
 			return;
 
 		sock = new QSslSocket(this);
-		connect(sock, SIGNAL(connected()), SLOT(sock_connected()));
-		connect(sock, SIGNAL(readyRead()), SLOT(sock_readyRead()));
-		connect(sock, SIGNAL(bytesWritten(qint64)), SLOT(sock_bytesWritten(qint64)));
-		connect(sock, SIGNAL(disconnected()), SLOT(sock_disconnected()));
-		connect(sock, SIGNAL(error(QAbstractSocket::SocketError)), SLOT(sock_error(QAbstractSocket::SocketError)));
-		connect(sock, SIGNAL(sslErrors(const QList<QSslError> &)), SLOT(sock_sslErrors(const QList<QSslError> &)));
+		connect(sock, &QSslSocket::connected, this, &Private::sock_connected);
+		connect(sock, &QSslSocket::readyRead, this, &Private::sock_readyRead);
+		connect(sock, &QSslSocket::bytesWritten, this, &Private::sock_bytesWritten);
+		connect(sock, &QSslSocket::disconnected, this, &Private::sock_disconnected);
+		connect(sock, static_cast<void (QSslSocket::*)(QAbstractSocket::SocketError)>(&QSslSocket::error), this, &Private::sock_error);
+		connect(sock, static_cast<void (QSslSocket::*)(const QList<QSslError> &)>(&QSslSocket::sslErrors), this, &Private::sock_sslErrors);
 
 		bool useSsl = (requestUri.scheme() == "wss");
 		int port = requestUri.port(useSsl ? 443 : 80);
@@ -871,19 +927,25 @@ private slots:
 
 		requestKey = generateKey();
 
-		requestHeaders.removeAll("Host");
 		requestHeaders.removeAll("Upgrade");
 		requestHeaders.removeAll("Connection");
 		requestHeaders.removeAll("Sec-WebSocket-Version");
 		requestHeaders.removeAll("Sec-WebSocket-Key");
-		requestHeaders.removeAll("Accept-Encoding"); // we only support unencoded rejections
+		requestHeaders.removeAll("Accept-Encoding"); // we only support unencoded responses
 
 		// note: we let Sec-WebSocket-Extensions and
 		//   Sec-WebSocket-Protocol go through. clients should take
 		//   care to not send connection-level extensions, as we won't
 		//   be able to understand them
 
-		requestHeaders += HttpHeader("Host", requestUri.host().toUtf8());
+		if(!requestHeaders.contains("Host"))
+		{
+			QByteArray hostHeader = requestUri.host().toUtf8();
+			if(requestUri.port() != -1)
+				hostHeader += ':' + QByteArray::number(requestUri.port());
+			requestHeaders += HttpHeader("Host", hostHeader);
+		}
+
 		requestHeaders += HttpHeader("Upgrade", "websocket");
 		requestHeaders += HttpHeader("Connection", "Upgrade");
 		requestHeaders += HttpHeader("Sec-WebSocket-Version", "13");
@@ -909,11 +971,11 @@ private slots:
 			log_debug("ws: read: %d", buf.size());
 			inbuf += buf;
 
-			if(!readingRejectBody)
+			if(!readingResponseBody)
 			{
 				QPointer<QObject> self = this;
 				bool ok = true;
-				while(state == Connecting && !readingRejectBody && ok)
+				while(state == Connecting && !readingResponseBody && ok)
 				{
 					int at = inbuf.indexOf('\n');
 					if(at == -1)
@@ -993,6 +1055,8 @@ private slots:
 	{
 		log_debug("ws: sock_error: %d", (int)socketError);
 
+		bool tryAgain = true;
+
 		ErrorCondition curError;
 		switch(socketError)
 		{
@@ -1000,21 +1064,22 @@ private slots:
 				curError = ErrorConnect;
 				break;
 			case QAbstractSocket::RemoteHostClosedError:
-				if(readingRejectBody && responseContentLength == -1 && !chunked)
+				if(readingResponseBody && responseContentLength == -1 && !chunked)
 				{
-					respondRejected();
+					handleResponse();
 					return;
 				}
 				curError = ErrorGeneric;
 				break;
 			case QAbstractSocket::SslHandshakeFailedError:
 				curError = ErrorTls;
+				tryAgain = false;
 				break;
 			default:
 				curError = ErrorGeneric;
 		}
 
-		if(state == Connected)
+		if(!tryAgain || state == Connected)
 		{
 			cleanup();
 			state = Idle;
@@ -1032,9 +1097,37 @@ private slots:
 
 	void sock_sslErrors(const QList<QSslError> &errors)
 	{
-		Q_UNUSED(errors);
+		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
+		{
+			QStringList strs;
+			foreach(const QSslError &e, errors)
+				strs += e.errorString();
+			log_debug("ws: sslErrors: %s", qPrintable(strs.join(", ")));
+		}
 
-		if(ignoreTlsErrors)
+		bool hostMismatchOk = false;
+		if(errors.count() == 1 && errors[0].error() == QSslError::HostNameMismatch)
+		{
+			// if hostname doesn't match, check against connect host if trusted
+			if(!connectHost.isEmpty() && trustConnectHost)
+			{
+#ifdef HAVE_OPENSSL
+				QSslCertificate cert = sock->peerCertificate();
+				QByteArray der = cert.toDer();
+				const unsigned char *p = (const unsigned char *)der.data();
+				X509 *opensslCert = d2i_X509(NULL, &p, der.size());
+				if(opensslCert)
+				{
+					if(verifyhost(connectHost.toUtf8().data(), opensslCert) == CURLE_OK)
+						hostMismatchOk = true;
+
+					X509_free(opensslCert);
+				}
+#endif
+			}
+		}
+
+		if(ignoreTlsErrors || hostMismatchOk)
 			sock->ignoreSslErrors();
 	}
 };
@@ -1055,11 +1148,21 @@ void WebSocket::setConnectHost(const QString &host)
 	d->connectHost = host;
 }
 
+void WebSocket::setTrustConnectHost(bool on)
+{
+	d->trustConnectHost = on;
+}
+
 void WebSocket::setIgnoreTlsErrors(bool on)
 {
 	d->ignoreTlsErrors = on;
 }
 
+void WebSocket::setFollowRedirects(int maxRedirects)
+{
+	d->maxRedirects = maxRedirects;
+}
+
 void WebSocket::setMaxFrameSize(int size)
 {
 	d->maxFrameSize = size;
