@@ -43,6 +43,19 @@ public:
 		WebSocketTransport
 	};
 
+	enum State
+	{
+		NotStarted,
+		Started,
+		Closing,
+		PeerClosing,
+		CloseWait,
+		Finished,
+		Cancel,
+		Error,
+		Stopped
+	};
+
 	Worker *q;
 	QJDnsShared *dns;
 	AppConfig *config;
@@ -50,6 +63,8 @@ public:
 	Worker::Format format;
 	QByteArray toAddress;
 	QByteArray rid;
+	State state;
+	QByteArray errorCondition;
 	int inSeq, outSeq;
 	int outCredits;
 	bool outStream;
@@ -65,15 +80,15 @@ public:
 	bool stuffToRead;
 	BufferList inbuf; // for single mode
 	int bytesReceived;
-	bool pendingSend;
 	QTimer *expireTimer;
 	QTimer *httpActivityTimer;
 	QTimer *httpSessionTimer;
 	QTimer *keepAliveTimer;
+	QTimer *updateTimer;
 	WebSocket::Frame::Type lastReceivedFrameType;
 	bool wsSendingMessage;
 	QList<int> wsPendingWrites;
-	bool wsPeerClosing;
+	bool wsClosed;
 	bool wsPendingPeerClose;
 
 	Private(QJDnsShared *_dns, AppConfig *_config, Worker::Format _format, Worker *_q) :
@@ -82,26 +97,36 @@ public:
 		dns(_dns),
 		config(_config),
 		format(_format),
+		state(NotStarted),
 		hreq(0),
 		ws(0),
-		pendingSend(false),
 		expireTimer(0),
 		httpActivityTimer(0),
 		httpSessionTimer(0),
 		keepAliveTimer(0),
 		lastReceivedFrameType(WebSocket::Frame::Text),
 		wsSendingMessage(false),
-		wsPeerClosing(false),
+		wsClosed(false),
 		wsPendingPeerClose(false)
 	{
+		updateTimer = new QTimer(this);
+		connect(updateTimer, &QTimer::timeout, this, &Private::doUpdate);
+		updateTimer->setSingleShot(true);
 	}
 
 	~Private()
 	{
+		cleanup();
+
+		updateTimer->disconnect(this);
+		updateTimer->setParent(0);
+		updateTimer->deleteLater();
 	}
 
 	void cleanup()
 	{
+		updateTimer->stop();
+
 		delete hreq;
 		hreq = 0;
 
@@ -139,6 +164,8 @@ public:
 			keepAliveTimer->deleteLater();
 			keepAliveTimer = 0;
 		}
+
+		state = Stopped;
 	}
 
 	void start(const QVariant &vrequest, Mode mode)
@@ -147,6 +174,8 @@ public:
 		outCredits = 0;
 		quiet = false;
 
+		state = Started;
+
 		ZhttpRequestPacket request;
 		if(!request.fromVariant(vrequest))
 		{
@@ -157,14 +186,9 @@ public:
 			toAddress = vhash.value("from").toByteArray();
 			QByteArray type = vhash.value("type").toByteArray();
 			if(!toAddress.isEmpty() && type != "error" && type != "cancel")
-			{
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
-			}
+				deferError("bad-request");
 			else
-			{
-				cleanup();
-				QMetaObject::invokeMethod(q, "finished", Qt::QueuedConnection);
-			}
+				deferFinished();
 
 			return;
 		}
@@ -177,12 +201,13 @@ public:
 		bytesReceived = 0;
 
 		ignorePolicies = request.ignorePolicies;
+		sessionTimeout = -1;
 
 		if(request.uri.isEmpty())
 		{
 			log_warning("missing request uri");
 
-			QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+			deferError("bad-request");
 			return;
 		}
 
@@ -199,7 +224,7 @@ public:
 		{
 			log_warning("unsupported scheme");
 
-			QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+			deferError("bad-request");
 			return;
 		}
 
@@ -207,7 +232,7 @@ public:
 		{
 			log_warning("websocket must be used from stream interface");
 
-			QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+			deferError("bad-request");
 			return;
 		}
 
@@ -235,7 +260,7 @@ public:
 			{
 				log_warning("missing request method");
 
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+				deferError("bad-request");
 				return;
 			}
 
@@ -246,7 +271,7 @@ public:
 			{
 				log_warning("streamed input must start with seq 0");
 
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+				deferError("bad-request");
 				return;
 			}
 
@@ -255,7 +280,7 @@ public:
 			{
 				log_warning("cannot use streamed input on router interface");
 
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+				deferError("bad-request");
 				return;
 			}
 
@@ -265,25 +290,27 @@ public:
 
 			if(!isAllowed(request.uri.host()) || (!request.connectHost.isEmpty() && !isAllowed(request.connectHost)))
 			{
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "policy-violation"));
+				deferError("policy-violation");
 				return;
 			}
 
 			QByteArray hostHeader = request.uri.host().toUtf8();
 
-			// only tack on the port if it isn't being overridden
 			int port = request.uri.port(defaultPort);
-			if(request.connectPort == -1 && port != defaultPort)
-				hostHeader += ":" + QByteArray::number(port);
+			if(port != defaultPort)
+				hostHeader += ':' + QByteArray::number(port);
 
-			headers.removeAll("Host");
-			headers += HttpHeader("Host", hostHeader);
+			if(headers.get("Host") != hostHeader)
+			{
+				headers.removeAll("Host");
+				headers += HttpHeader("Host", hostHeader);
+			}
 
 			hreq = new HttpRequest(dns, this);
-			connect(hreq, SIGNAL(nextAddress(const QHostAddress &)), SLOT(req_nextAddress(const QHostAddress &)));
-			connect(hreq, SIGNAL(readyRead()), SLOT(req_readyRead()));
-			connect(hreq, SIGNAL(bytesWritten(int)), SLOT(req_bytesWritten(int)));
-			connect(hreq, SIGNAL(error()), SLOT(req_error()));
+			connect(hreq, &HttpRequest::nextAddress, this, &Private::req_nextAddress);
+			connect(hreq, &HttpRequest::readyRead, this, &Private::req_readyRead);
+			connect(hreq, &HttpRequest::bytesWritten, this, &Private::req_bytesWritten);
+			connect(hreq, &HttpRequest::error, this, &Private::req_error);
 
 			maxResponseSize = request.maxSize;
 			sessionTimeout = request.timeout;
@@ -293,6 +320,7 @@ public:
 			if(request.connectPort != -1)
 				request.uri.setPort(request.connectPort);
 
+			hreq->setTrustConnectHost(request.trustConnectHost);
 			hreq->setIgnoreTlsErrors(request.ignoreTlsErrors);
 			if(request.followRedirects)
 				hreq->setFollowRedirects(8);
@@ -309,7 +337,7 @@ public:
 			{
 				log_warning("websocket input must start with seq 0");
 
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+				deferError("bad-request");
 				return;
 			}
 
@@ -317,7 +345,7 @@ public:
 			{
 				log_warning("websocket input must provide from address");
 
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+				deferError("bad-request");
 				return;
 			}
 
@@ -325,35 +353,40 @@ public:
 
 			if(!isAllowed(request.uri.host()) || (!request.connectHost.isEmpty() && !isAllowed(request.connectHost)))
 			{
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "policy-violation"));
+				deferError("policy-violation");
 				return;
 			}
 
 			QByteArray hostHeader = request.uri.host().toUtf8();
 
-			// only tack on the port if it isn't being overridden
 			int port = request.uri.port(defaultPort);
-			if(request.connectPort == -1 && port != defaultPort)
+			if(port != defaultPort)
 				hostHeader += ":" + QByteArray::number(port);
 
-			headers.removeAll("Host");
-			headers += HttpHeader("Host", hostHeader);
+			if(headers.get("Host") != hostHeader)
+			{
+				headers.removeAll("Host");
+				headers += HttpHeader("Host", hostHeader);
+			}
 
 			ws = new WebSocket(dns, this);
-			connect(ws, SIGNAL(nextAddress(const QHostAddress &)), SLOT(req_nextAddress(const QHostAddress &)));
-			connect(ws, SIGNAL(connected()), SLOT(ws_connected()));
-			connect(ws, SIGNAL(readyRead()), SLOT(ws_readyRead()));
-			connect(ws, SIGNAL(framesWritten(int)), SLOT(ws_framesWritten(int)));
-			connect(ws, SIGNAL(peerClosing()), SLOT(ws_peerClosing()));
-			connect(ws, SIGNAL(closed()), SLOT(ws_closed()));
-			connect(ws, SIGNAL(error()), SLOT(ws_error()));
+			connect(ws, &WebSocket::nextAddress, this, &Private::req_nextAddress);
+			connect(ws, &WebSocket::connected, this, &Private::ws_connected);
+			connect(ws, &WebSocket::readyRead, this, &Private::ws_readyRead);
+			connect(ws, &WebSocket::framesWritten, this, &Private::ws_framesWritten);
+			connect(ws, &WebSocket::peerClosing, this, &Private::ws_peerClosing);
+			connect(ws, &WebSocket::closed, this, &Private::ws_closed);
+			connect(ws, &WebSocket::error, this, &Private::ws_error);
 
 			if(!request.connectHost.isEmpty())
 				ws->setConnectHost(request.connectHost);
 			if(request.connectPort != -1)
 				request.uri.setPort(request.connectPort);
 
+			ws->setTrustConnectHost(request.trustConnectHost);
 			ws->setIgnoreTlsErrors(request.ignoreTlsErrors);
+			if(request.followRedirects)
+				ws->setFollowRedirects(8);
 			ws->setMaxFrameSize(config->sessionBufferSize);
 
 			if(request.credits != -1)
@@ -361,14 +394,14 @@ public:
 		}
 
 		httpActivityTimer = new QTimer(this);
-		connect(httpActivityTimer, SIGNAL(timeout()), SLOT(httpActivity_timeout()));
+		connect(httpActivityTimer, &QTimer::timeout, this, &Private::httpActivity_timeout);
 		httpActivityTimer->setSingleShot(true);
 		httpActivityTimer->start(config->activityTimeout * 1000);
 
 		if(sessionTimeout != -1)
 		{
 			httpSessionTimer = new QTimer(this);
-			connect(httpSessionTimer, SIGNAL(timeout()), SLOT(httpSession_timeout()));
+			connect(httpSessionTimer, &QTimer::timeout, this, &Private::httpSession_timeout);
 			httpSessionTimer->setSingleShot(true);
 			httpSessionTimer->start(sessionTimeout);
 		}
@@ -376,12 +409,12 @@ public:
 		if(transport == WebSocketTransport || (transport == HttpTransport && mode == Worker::Stream))
 		{
 			expireTimer = new QTimer(this);
-			connect(expireTimer, SIGNAL(timeout()), SLOT(expire_timeout()));
+			connect(expireTimer, &QTimer::timeout, this, &Private::expire_timeout);
 			expireTimer->setSingleShot(true);
 			expireTimer->start(SESSION_EXPIRE);
 
 			keepAliveTimer = new QTimer(this);
-			connect(keepAliveTimer, SIGNAL(timeout()), SLOT(keepAlive_timeout()));
+			connect(keepAliveTimer, &QTimer::timeout, this, &Private::keepAlive_timeout);
 			keepAliveTimer->start(SESSION_EXPIRE / 2);
 		}
 
@@ -435,19 +468,17 @@ public:
 
 	void write(const QVariant &vrequest)
 	{
+		if(!(state == Started || state == Closing || state == PeerClosing))
+			return;
+
 		ZhttpRequestPacket request;
 		if(!request.fromVariant(vrequest))
 		{
 			QVariantHash vhash = vrequest.toHash();
 			if(vhash["type"].toByteArray() != "cancel")
-			{
-				QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
-			}
+				deferError("bad-request");
 			else
-			{
-				cleanup();
-				QMetaObject::invokeMethod(q, "finished", Qt::QueuedConnection);
-			}
+				deferFinished();
 
 			return;
 		}
@@ -456,22 +487,16 @@ public:
 		if(inSeq == -1 || request.seq == -1 || request.seq != inSeq + 1)
 		{
 			if(request.type != ZhttpRequestPacket::Cancel)
-			{
-				QMetaObject::invokeMethod(this, "respondCancel", Qt::QueuedConnection);
-			}
+				deferCancel();
 			else
-			{
-				cleanup();
-				QMetaObject::invokeMethod(q, "finished", Qt::QueuedConnection);
-			}
+				deferFinished();
 
 			return;
 		}
 
 		if(request.type == ZhttpRequestPacket::Cancel)
 		{
-			cleanup();
-			QMetaObject::invokeMethod(q, "finished", Qt::QueuedConnection);
+			deferFinished();
 			return;
 		}
 
@@ -490,7 +515,7 @@ public:
 			{
 				if(bodySent)
 				{
-					QMetaObject::invokeMethod(this, "respondError", Qt::QueuedConnection, Q_ARG(QByteArray, "bad-request"));
+					deferError("bad-request");
 					return;
 				}
 
@@ -511,41 +536,59 @@ public:
 		{
 			if(request.type == ZhttpRequestPacket::Data || request.type == ZhttpRequestPacket::Close || request.type == ZhttpRequestPacket::Ping || request.type == ZhttpRequestPacket::Pong)
 			{
-				refreshActivityTimeout();
-
-				if(request.type == ZhttpRequestPacket::Data)
+				if(wsClosed)
 				{
-					WebSocket::Frame::Type ftype;
-					if(wsSendingMessage)
-						ftype = WebSocket::Frame::Continuation;
-					else if(request.contentType == "binary")
-						ftype = WebSocket::Frame::Binary;
-					else
-						ftype = WebSocket::Frame::Text;
-
-					wsSendingMessage = request.more;
-
-					wsPendingWrites += request.body.size();
-					ws->writeFrame(WebSocket::Frame(ftype, request.body, request.more));
+					deferError("bad-request");
+					return;
 				}
-				else if(request.type == ZhttpRequestPacket::Ping)
+
+				if(request.type == ZhttpRequestPacket::Data || request.type == ZhttpRequestPacket::Ping || request.type == ZhttpRequestPacket::Pong)
 				{
-					wsPendingWrites += 0;
-					ws->writeFrame(WebSocket::Frame(WebSocket::Frame::Ping, QByteArray(), false));
-				}
-				else if(request.type == ZhttpRequestPacket::Pong)
-				{
-					wsPendingWrites += 0;
-					ws->writeFrame(WebSocket::Frame(WebSocket::Frame::Pong, QByteArray(), false));
+					refreshActivityTimeout();
+
+					if(request.type == ZhttpRequestPacket::Data)
+					{
+						WebSocket::Frame::Type ftype;
+						if(wsSendingMessage)
+							ftype = WebSocket::Frame::Continuation;
+						else if(request.contentType == "binary")
+							ftype = WebSocket::Frame::Binary;
+						else
+							ftype = WebSocket::Frame::Text;
+
+						wsSendingMessage = request.more;
+
+						wsPendingWrites += request.body.size();
+						ws->writeFrame(WebSocket::Frame(ftype, request.body, request.more));
+					}
+					else if(request.type == ZhttpRequestPacket::Ping)
+					{
+						wsPendingWrites += 0;
+						ws->writeFrame(WebSocket::Frame(WebSocket::Frame::Ping, QByteArray(), false));
+					}
+					else if(request.type == ZhttpRequestPacket::Pong)
+					{
+						wsPendingWrites += 0;
+						ws->writeFrame(WebSocket::Frame(WebSocket::Frame::Pong, QByteArray(), false));
+					}
 				}
 				else if(request.type == ZhttpRequestPacket::Close)
+				{
 					ws->close(request.code);
+
+					wsClosed = true;
+
+					if(state == Started)
+						state = Closing;
+					else // PeerClosing
+						state = CloseWait;
+				}
 			}
 		}
 
 		// if we needed credits to send something, take care of that now
 		if(request.credits != -1 && stuffToRead)
-			trySend();
+			update();
 	}
 
 	static bool matchExp(const QString &exp, const QString &s)
@@ -623,13 +666,32 @@ public:
 			emit q->readyRead(toAddress, out.toVariant());
 	}
 
-	void doSend()
+	void update()
 	{
-		if(!pendingSend)
-		{
-			pendingSend = true;
-			QMetaObject::invokeMethod(this, "trySend", Qt::QueuedConnection);
-		}
+		if(!updateTimer->isActive())
+			updateTimer->start();
+	}
+
+	void deferFinished()
+	{
+		cleanup();
+		state = Finished;
+		update();
+	}
+
+	void deferCancel()
+	{
+		cleanup();
+		state = Cancel;
+		update();
+	}
+
+	void deferError(const QByteArray &condition)
+	{
+		cleanup();
+		state = Error;
+		errorCondition = condition;
+		update();
 	}
 
 	void refreshTimeout()
@@ -642,151 +704,6 @@ public:
 		httpActivityTimer->start(config->activityTimeout * 1000);
 	}
 
-	void tryCleanup()
-	{
-		if(ws->state() == WebSocket::Idle && ws->framesAvailable() == 0)
-		{
-			cleanup();
-			emit q->finished();
-		}
-	}
-
-private slots:
-	// emits signals, but safe to delete after
-	void trySend()
-	{
-		QPointer<QObject> self = this;
-
-		pendingSend = false;
-		stuffToRead = false;
-
-		if(transport == HttpTransport)
-		{
-			ZhttpResponsePacket resp;
-			resp.type = ZhttpResponsePacket::Data;
-
-			if(!sentHeader)
-			{
-				resp.code = hreq->responseCode();
-				resp.reason = hreq->responseReason();
-				resp.headers = hreq->responseHeaders();
-				sentHeader = true;
-			}
-
-			// note: we always set body, even if empty
-
-			if(outStream)
-			{
-				// note: we skip credits handling if quiet mode
-
-				QByteArray buf;
-
-				if(!quiet)
-					buf = hreq->readResponseBody(outCredits);
-				else
-					buf = hreq->readResponseBody(); // all
-
-				if(!buf.isEmpty())
-				{
-					if(maxResponseSize != -1 && bytesReceived + buf.size() > maxResponseSize)
-					{
-						respondError("max-size-exceeded");
-						return;
-					}
-				}
-
-				resp.body = buf;
-
-				bytesReceived += resp.body.size();
-
-				if(!quiet)
-					outCredits -= resp.body.size();
-
-				resp.more = (hreq->bytesAvailable() > 0 || !hreq->isFinished());
-
-				if(hreq->bytesAvailable() > 0)
-					stuffToRead = true;
-			}
-			else
-			{
-				resp.body = inbuf.take();
-			}
-
-			writeResponse(resp);
-			if(!self)
-				return;
-
-			if(!resp.more)
-			{
-				cleanup();
-				emit q->finished();
-			}
-		}
-		else // WebSocketTransport
-		{
-			while(ws->framesAvailable() > 0 && outCredits >= ws->nextFrameSize())
-			{
-				WebSocket::Frame frame = ws->readFrame();
-				outCredits -= frame.data.size();
-
-				if(frame.type == WebSocket::Frame::Continuation || frame.type == WebSocket::Frame::Text || frame.type == WebSocket::Frame::Binary)
-				{
-					if(frame.type == WebSocket::Frame::Continuation)
-						frame.type = lastReceivedFrameType;
-					else
-						lastReceivedFrameType = frame.type;
-
-					ZhttpResponsePacket resp;
-					resp.type = ZhttpResponsePacket::Data;
-					if(frame.type == WebSocket::Frame::Binary)
-						resp.contentType = "binary";
-					resp.body = frame.data;
-					resp.more = frame.more;
-					writeResponse(resp);
-					if(!self)
-						return;
-				}
-				else if(frame.type == WebSocket::Frame::Ping)
-				{
-					ZhttpResponsePacket resp;
-					resp.type = ZhttpResponsePacket::Ping;
-					writeResponse(resp);
-					if(!self)
-						return;
-				}
-				else if(frame.type == WebSocket::Frame::Pong)
-				{
-					ZhttpResponsePacket resp;
-					resp.type = ZhttpResponsePacket::Pong;
-					writeResponse(resp);
-					if(!self)
-						return;
-				}
-			}
-
-			if(ws->framesAvailable() > 0)
-			{
-				stuffToRead = true;
-			}
-			else
-			{
-				if(wsPendingPeerClose)
-				{
-					wsPendingPeerClose = false;
-
-					ZhttpResponsePacket resp;
-					resp.type = ZhttpResponsePacket::Close;
-					resp.code = ws->peerCloseCode();
-					writeResponse(resp);
-					if(!self)
-						return;
-				}
-
-				tryCleanup();
-			}
-		}
-	}
-
 	void respondError(const QByteArray &condition)
 	{
 		respondError(condition, -1, QByteArray(), HttpHeaders(), QByteArray());
@@ -831,6 +748,190 @@ private slots:
 		emit q->finished();
 	}
 
+private slots:
+	// emits signals, but safe to delete after
+	void doUpdate()
+	{
+		QPointer<QObject> self = this;
+
+		// if we had a pending update, we can cancel since we're updating now
+		if(updateTimer->isActive())
+			updateTimer->stop();
+
+		if(transport == HttpTransport)
+		{
+			if(state == Started)
+			{
+				if(!stuffToRead)
+					return;
+
+				stuffToRead = false;
+
+				ZhttpResponsePacket resp;
+				resp.type = ZhttpResponsePacket::Data;
+
+				if(!sentHeader)
+				{
+					resp.code = hreq->responseCode();
+					resp.reason = hreq->responseReason();
+					resp.headers = hreq->responseHeaders();
+					sentHeader = true;
+				}
+
+				// note: we always set body, even if empty
+
+				if(outStream)
+				{
+					// note: we skip credits handling if quiet mode
+
+					QByteArray buf;
+
+					if(!quiet)
+						buf = hreq->readResponseBody(outCredits);
+					else
+						buf = hreq->readResponseBody(); // all
+
+					if(!buf.isEmpty())
+					{
+						if(maxResponseSize != -1 && bytesReceived + buf.size() > maxResponseSize)
+						{
+							respondError("max-size-exceeded");
+							return;
+						}
+					}
+
+					resp.body = buf;
+
+					bytesReceived += resp.body.size();
+
+					if(!quiet)
+						outCredits -= resp.body.size();
+
+					resp.more = (hreq->bytesAvailable() > 0 || !hreq->isFinished());
+
+					if(hreq->bytesAvailable() > 0)
+						stuffToRead = true;
+				}
+				else
+				{
+					resp.body = inbuf.take();
+				}
+
+				writeResponse(resp);
+				if(!self)
+					return;
+
+				if(!resp.more)
+				{
+					cleanup();
+					emit q->finished();
+				}
+			}
+			else if(state == Finished)
+			{
+				cleanup();
+				emit q->finished();
+			}
+			else if(state == Cancel)
+			{
+				respondCancel();
+			}
+			else if(state == Error)
+			{
+				respondError(errorCondition);
+			}
+		}
+		else // WebSocketTransport
+		{
+			if(state == Started || state || Closing || state == PeerClosing)
+			{
+				if(stuffToRead)
+				{
+					stuffToRead = false;
+
+					while(ws->framesAvailable() > 0 && outCredits >= ws->nextFrameSize())
+					{
+						WebSocket::Frame frame = ws->readFrame();
+						outCredits -= frame.data.size();
+
+						if(frame.type == WebSocket::Frame::Continuation || frame.type == WebSocket::Frame::Text || frame.type == WebSocket::Frame::Binary)
+						{
+							if(frame.type == WebSocket::Frame::Continuation)
+								frame.type = lastReceivedFrameType;
+							else
+								lastReceivedFrameType = frame.type;
+
+							ZhttpResponsePacket resp;
+							resp.type = ZhttpResponsePacket::Data;
+							if(frame.type == WebSocket::Frame::Binary)
+								resp.contentType = "binary";
+							resp.body = frame.data;
+							resp.more = frame.more;
+							writeResponse(resp);
+							if(!self)
+								return;
+						}
+						else if(frame.type == WebSocket::Frame::Ping)
+						{
+							ZhttpResponsePacket resp;
+							resp.type = ZhttpResponsePacket::Ping;
+							writeResponse(resp);
+							if(!self)
+								return;
+						}
+						else if(frame.type == WebSocket::Frame::Pong)
+						{
+							ZhttpResponsePacket resp;
+							resp.type = ZhttpResponsePacket::Pong;
+							writeResponse(resp);
+							if(!self)
+								return;
+						}
+					}
+
+					if(ws->framesAvailable() > 0)
+					{
+						stuffToRead = true;
+					}
+				}
+
+				if(wsPendingPeerClose && !stuffToRead)
+				{
+					wsPendingPeerClose = false;
+
+					ZhttpResponsePacket resp;
+					resp.type = ZhttpResponsePacket::Close;
+					resp.code = ws->peerCloseCode();
+					writeResponse(resp);
+					if(!self)
+						return;
+
+					if(state == Closing)
+					{
+						cleanup();
+						emit q->finished();
+						return;
+					}
+
+					state = PeerClosing;
+				}
+			}
+			else if(state == Finished)
+			{
+				cleanup();
+				emit q->finished();
+			}
+			else if(state == Cancel)
+			{
+				respondCancel();
+			}
+			else if(state == Error)
+			{
+				respondError(errorCondition);
+			}
+		}
+	}
+
 	void req_nextAddress(const QHostAddress &addr)
 	{
 		if(!isAllowed(addr.toString()))
@@ -870,7 +971,7 @@ private slots:
 		}
 
 		// this will defer for a cycle, to pick up a potential disconnect as well
-		doSend();
+		update();
 	}
 
 	void req_bytesWritten(int count)
@@ -932,7 +1033,7 @@ private slots:
 		if(outCredits < 1)
 			return;
 
-		trySend();
+		doUpdate();
 	}
 
 	void ws_framesWritten(int count)
@@ -949,25 +1050,28 @@ private slots:
 
 	void ws_peerClosing()
 	{
-		wsPeerClosing = true;
+		assert(state == Started);
+
+		// flag to peer close once all queued data is sent
 		wsPendingPeerClose = true;
 
-		trySend();
+		doUpdate();
 	}
 
 	void ws_closed()
 	{
-		if(wsPeerClosing)
+		if(state == Closing)
 		{
-			// we were acking the peer's close, so we're done. well, as soon as all
-			//   data is read that is.
-			tryCleanup();
-		}
-		else
-		{
-			// the peer is acking our close. flag to send along the ack
+			// flag to peer close once all queued data is sent
 			wsPendingPeerClose = true;
-			trySend();
+
+			doUpdate();
+		}
+		else if(state == CloseWait)
+		{
+			// we were acking the peer's close, so we're done
+			cleanup();
+			emit q->finished();
 		}
 	}
 
